@@ -1,8 +1,9 @@
 /*
- *  Copyright (C) 2013-2018 Savoir-faire Linux Inc.
+ *  Copyright (C) 2013-2019 Savoir-faire Linux Inc.
  *
  *  Author: Guillaume Roguez <Guillaume.Roguez@savoirfairelinux.com>
  *  Author: Eloi Bail <Eloi.Bail@savoirfairelinux.com>
+ *  Author: Philippe Gorley <philippe.gorley@savoirfairelinux.com>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -23,19 +24,26 @@
 #include "media_codec.h"
 #include "media_encoder.h"
 #include "media_buffer.h"
-#include "media_io_handle.h"
 
-#include "audio/audiobuffer.h"
-#include "string_utils.h"
+#include "client/ring_signal.h"
+#include "fileutils.h"
 #include "logger.h"
+#include "manager.h"
+#include "string_utils.h"
+
+#ifdef RING_ACCEL
+#include "video/accel.h"
+#endif
 
 extern "C" {
 #include <libavutil/parseutils.h>
 }
 
-#include <iostream>
-#include <sstream>
 #include <algorithm>
+#include <fstream>
+#include <iostream>
+#include <json/json.h>
+#include <sstream>
 #include <thread> // hardware_concurrency
 
 // Define following line if you need to debug libav SDP
@@ -52,8 +60,7 @@ MediaEncoder::~MediaEncoder()
     if (outputCtx_) {
         if (outputCtx_->priv_data)
             av_write_trailer(outputCtx_);
-
-        for (auto encoderCtx : encoders_)
+        for (auto encoderCtx : encoders_) {
             if (encoderCtx) {
 #ifndef _MSC_VER
                 avcodec_free_context(&encoderCtx);
@@ -61,112 +68,66 @@ MediaEncoder::~MediaEncoder()
                 avcodec_close(encoderCtx);
 #endif
             }
-
+        }
         avformat_free_context(outputCtx_);
     }
-
     av_dict_free(&options_);
 }
 
-void MediaEncoder::setDeviceOptions(const DeviceParams& args)
+void
+MediaEncoder::setDeviceOptions(const DeviceParams& args)
 {
     device_ = args;
+    // Make sure width and height are even (required by x264)
+    // This is especially for image/gif streaming, as video files and cameras usually have even resolutions
+    device_.width -= device_.width % 2;
+    device_.height -= device_.height % 2;
     if (device_.width)
-        av_dict_set(&options_, "width", ring::to_string(device_.width).c_str(), 0);
+        libav_utils::setDictValue(&options_, "width", ring::to_string(device_.width));
     if (device_.height)
-        av_dict_set(&options_, "height", ring::to_string(device_.height).c_str(), 0);
+        libav_utils::setDictValue(&options_, "height", ring::to_string(device_.height));
     if (not device_.framerate)
         device_.framerate = 30;
-    av_dict_set(&options_, "framerate", ring::to_string(device_.framerate.real()).c_str(), 0);
+    libav_utils::setDictValue(&options_, "framerate", ring::to_string(device_.framerate.real()));
 }
 
-void MediaEncoder::setOptions(const MediaDescription& args)
+void
+MediaEncoder::setOptions(const MediaDescription& args)
 {
     codec_ = args.codec;
 
-    av_dict_set(&options_, "payload_type", ring::to_string(args.payload_type).c_str(), 0);
-    av_dict_set(&options_, "max_rate", ring::to_string(args.codec->bitrate).c_str(), 0);
-    av_dict_set(&options_, "crf", ring::to_string(args.codec->quality).c_str(), 0);
+    libav_utils::setDictValue(&options_, "payload_type", ring::to_string(args.payload_type));
+    libav_utils::setDictValue(&options_, "max_rate", ring::to_string(args.codec->bitrate));
+    libav_utils::setDictValue(&options_, "crf", ring::to_string(args.codec->quality));
 
     if (args.codec->systemCodecInfo.mediaType == MEDIA_AUDIO) {
         auto accountAudioCodec = std::static_pointer_cast<AccountAudioCodecInfo>(args.codec);
         if (accountAudioCodec->audioformat.sample_rate)
-            av_dict_set(&options_, "sample_rate",
-                        ring::to_string(accountAudioCodec->audioformat.sample_rate).c_str(), 0);
+            libav_utils::setDictValue(&options_, "sample_rate",
+                        ring::to_string(accountAudioCodec->audioformat.sample_rate));
 
         if (accountAudioCodec->audioformat.nb_channels)
-            av_dict_set(&options_, "channels",
-                        ring::to_string(accountAudioCodec->audioformat.nb_channels).c_str(), 0);
+            libav_utils::setDictValue(&options_, "channels",
+                        ring::to_string(accountAudioCodec->audioformat.nb_channels));
 
         if (accountAudioCodec->audioformat.sample_rate && accountAudioCodec->audioformat.nb_channels)
-            av_dict_set(&options_, "frame_size",
-                        ring::to_string(static_cast<unsigned>(0.02 * accountAudioCodec->audioformat.sample_rate)).c_str(), 0);
+            libav_utils::setDictValue(&options_, "frame_size",
+                        ring::to_string(static_cast<unsigned>(0.02 * accountAudioCodec->audioformat.sample_rate)));
     }
 
     if (not args.parameters.empty())
-        av_dict_set(&options_, "parameters", args.parameters.c_str(), 0);
+        libav_utils::setDictValue(&options_, "parameters", args.parameters);
 }
 
 void
-MediaEncoder::setInitSeqVal(uint16_t seqVal)
+MediaEncoder::setOptions(std::map<std::string, std::string> options)
 {
-    //only set not default value (!=0)
-    if (seqVal != 0)
-        av_dict_set(&options_, "seq", ring::to_string(seqVal).c_str(), 0);
-}
-
-uint16_t
-MediaEncoder::getLastSeqValue()
-{
-    int64_t  retVal;
-    auto ret = av_opt_get_int(outputCtx_->priv_data, "seq", AV_OPT_SEARCH_CHILDREN, &retVal);
-    if (ret == 0)
-        return (uint16_t) retVal;
-    else
-        return 0;
-}
-
-std::string
-MediaEncoder::getEncoderName() const
-{
-    return encoders_[currentStreamIdx_]->codec->name;
-}
-
-void
-MediaEncoder::openLiveOutput(const std::string& filename,
-                             const ring::MediaDescription& args)
-{
-    setOptions(args);
-    AVOutputFormat *oformat = av_guess_format("rtp", filename.c_str(), nullptr);
-
-    if (!oformat) {
-        RING_ERR("Unable to find a suitable output format for %s", filename.c_str());
-        throw MediaEncoderException("No output format");
-    }
-
-    outputCtx_->oformat = oformat;
-#if LIBAVFORMAT_VERSION_INT >= AV_VERSION_INT(58, 7, 100)
-    // c_str guarantees null termination
-    outputCtx_->url = av_strdup(filename.c_str()); // must be compatible with av_free
-#else
-    strncpy(outputCtx_->filename, filename.c_str(), sizeof(outputCtx_->filename));
-    // in case our filename is longer than the space reserved for AVFormatContext.filename
-    outputCtx_->filename[sizeof(outputCtx_->filename) - 1] = '\0';
-#endif
-
-    addStream(args.codec->systemCodecInfo, args.parameters);
-}
-
-void
-MediaEncoder::openFileOutput(const std::string& filename, std::map<std::string, std::string> options)
-{
-    avformat_free_context(outputCtx_);
-    avformat_alloc_output_context2(&outputCtx_, nullptr, nullptr, filename.c_str());
-
-    if (!options["title"].empty())
-        av_dict_set(&outputCtx_->metadata, "title", options["title"].c_str(), 0);
-    if (!options["description"].empty())
-        av_dict_set(&outputCtx_->metadata, "description", options["description"].c_str(), 0);
+    const auto& titleIt = options.find("title");
+    if (titleIt != options.end() and not titleIt->second.empty())
+        libav_utils::setDictValue(&outputCtx_->metadata, titleIt->first, titleIt->second);
+    const auto& descIt = options.find("description");
+    if (descIt != options.end() and not descIt->second.empty())
+        libav_utils::setDictValue(&outputCtx_->metadata, descIt->first, descIt->second);
 
     auto bitrate = SystemCodecInfo::DEFAULT_MAX_BITRATE;
     auto quality = SystemCodecInfo::DEFAULT_CODEC_QUALITY;
@@ -181,37 +142,104 @@ MediaEncoder::openFileOutput(const std::string& filename, std::map<std::string,
     options.insert({"height", "240"});
     options.insert({"framerate", "30.00"});
     for (const auto& it : options)
-        av_dict_set(&options_, it.first.c_str(), it.second.c_str(), 0);
-    // for a file output, addStream is done by the caller, as there may be multiple streams
+        libav_utils::setDictValue(&options_, it.first, it.second);
+}
+
+void
+MediaEncoder::setInitSeqVal(uint16_t seqVal)
+{
+    //only set not default value (!=0)
+    if (seqVal != 0)
+        av_opt_set_int(outputCtx_, "seq", seqVal, AV_OPT_SEARCH_CHILDREN);
+}
+
+uint16_t
+MediaEncoder::getLastSeqValue()
+{
+    int64_t retVal;
+    if (av_opt_get_int(outputCtx_, "seq", AV_OPT_SEARCH_CHILDREN, &retVal) >= 0)
+        return (uint16_t)retVal;
+    else
+        return 0;
+}
+
+std::string
+MediaEncoder::getEncoderName() const
+{
+    return encoders_[currentStreamIdx_]->codec->name;
+}
+
+void
+MediaEncoder::openOutput(const std::string& filename, const std::string& format)
+{
+    avformat_free_context(outputCtx_);
+    if (format.empty())
+        avformat_alloc_output_context2(&outputCtx_, nullptr, nullptr, filename.c_str());
+    else
+        avformat_alloc_output_context2(&outputCtx_, nullptr, format.c_str(), filename.c_str());
+
+#ifdef RING_ACCEL
+    enableAccel_ = Manager::instance().videoPreferences.getEncodingAccelerated();
+#endif
 }
 
 int
-MediaEncoder::addStream(const SystemCodecInfo& systemCodecInfo, std::string parameters)
+MediaEncoder::addStream(const SystemCodecInfo& systemCodecInfo)
 {
     AVCodec* outputCodec = nullptr;
     AVCodecContext* encoderCtx = nullptr;
-    /* find the video encoder */
-    if (systemCodecInfo.avcodecId == AV_CODEC_ID_H263)
-        // For H263 encoding, we force the use of AV_CODEC_ID_H263P (H263-1998)
-        // H263-1998 can manage all frame sizes while H263 don't
-        // AV_CODEC_ID_H263 decoder will be used for decoding
-        outputCodec = avcodec_find_encoder(AV_CODEC_ID_H263P);
-    else
-        outputCodec = avcodec_find_encoder(static_cast<AVCodecID>(systemCodecInfo.avcodecId));
+#ifdef RING_ACCEL
+    if (systemCodecInfo.mediaType == MEDIA_VIDEO) {
+        if (enableAccel_) {
+            if (accel_ = video::HardwareAccel::setupEncoder(
+                static_cast<AVCodecID>(systemCodecInfo.avcodecId), device_.width, device_.height)) {
+                outputCodec = avcodec_find_encoder_by_name(accel_->getCodecName().c_str());
+            }
+        } else {
+            RING_WARN() << "Hardware encoding disabled";
+        }
+    }
+#endif
+
     if (!outputCodec) {
-        RING_ERR("Encoder \"%s\" not found!", systemCodecInfo.name.c_str());
-        throw MediaEncoderException("No output encoder");
+        /* find the video encoder */
+        if (systemCodecInfo.avcodecId == AV_CODEC_ID_H263)
+            // For H263 encoding, we force the use of AV_CODEC_ID_H263P (H263-1998)
+            // H263-1998 can manage all frame sizes while H263 don't
+            // AV_CODEC_ID_H263 decoder will be used for decoding
+            outputCodec = avcodec_find_encoder(AV_CODEC_ID_H263P);
+        else
+            outputCodec = avcodec_find_encoder(static_cast<AVCodecID>(systemCodecInfo.avcodecId));
+        if (!outputCodec) {
+            RING_ERR("Encoder \"%s\" not found!", systemCodecInfo.name.c_str());
+            throw MediaEncoderException("No output encoder");
+        }
     }
 
     encoderCtx = prepareEncoderContext(outputCodec, systemCodecInfo.mediaType == MEDIA_VIDEO);
     encoders_.push_back(encoderCtx);
+
+#ifdef RING_ACCEL
+    if (accel_) {
+        accel_->setDetails(encoderCtx, &options_);
+        encoderCtx->opaque = accel_.get();
+    }
+#endif
+
     auto maxBitrate = 1000 * std::atoi(libav_utils::getDictValue(options_, "max_rate"));
     auto bufSize = 2 * maxBitrate; // as recommended (TODO: make it customizable)
     auto crf = std::atoi(libav_utils::getDictValue(options_, "crf"));
 
     /* let x264 preset override our encoder settings */
     if (systemCodecInfo.avcodecId == AV_CODEC_ID_H264) {
-        extractProfileLevelID(parameters, encoderCtx);
+        auto profileLevelId = libav_utils::getDictValue(options_, "parameters");
+        extractProfileLevelID(profileLevelId, encoderCtx);
+#ifdef RING_ACCEL
+        if (accel_)
+            // limit the bitrate else it will easily go up to a few MiB/s
+            encoderCtx->bit_rate = maxBitrate;
+        else
+#endif
         forcePresetX264(encoderCtx);
         // For H264 :
         // Streaming => VBV (constrained encoding) + CRF (Constant Rate Factor)
@@ -219,7 +247,7 @@ MediaEncoder::addStream(const SystemCodecInfo& systemCodecInfo, std::string para
             crf = 30; // good value for H264-720p@30
         RING_DBG("H264 encoder setup: crf=%u, maxrate=%u, bufsize=%u", crf, maxBitrate, bufSize);
 
-        av_opt_set_int(encoderCtx->priv_data, "crf", crf, 0);
+        av_opt_set_int(encoderCtx, "crf", crf, AV_OPT_SEARCH_CHILDREN);
         encoderCtx->rc_buffer_size = bufSize;
         encoderCtx->rc_max_rate = maxBitrate;
     } else if (systemCodecInfo.avcodecId == AV_CODEC_ID_VP8) {
@@ -229,14 +257,14 @@ MediaEncoder::addStream(const SystemCodecInfo& systemCodecInfo, std::string para
         // 2- otherwise set rc_max_rate and rc_buffer_size
         // Using information given on this page:
         // http://www.webmproject.org/docs/encoder-parameters/
-        av_opt_set(encoderCtx->priv_data, "quality", "realtime", 0);
-        av_opt_set_int(encoderCtx->priv_data, "error-resilient", 1, 0);
-        av_opt_set_int(encoderCtx->priv_data, "cpu-used", 7, 0); // value obtained from testing
-        av_opt_set_int(encoderCtx->priv_data, "lag-in-frames", 0, 0);
+        av_opt_set(encoderCtx, "quality", "realtime", AV_OPT_SEARCH_CHILDREN);
+        av_opt_set_int(encoderCtx, "error-resilient", 1, AV_OPT_SEARCH_CHILDREN);
+        av_opt_set_int(encoderCtx, "cpu-used", 7, AV_OPT_SEARCH_CHILDREN); // value obtained from testing
+        av_opt_set_int(encoderCtx, "lag-in-frames", 0, AV_OPT_SEARCH_CHILDREN);
         // allow encoder to drop frames if buffers are full and
         // to undershoot target bitrate to lessen strain on resources
-        av_opt_set_int(encoderCtx->priv_data, "drop-frame", 25, 0);
-        av_opt_set_int(encoderCtx->priv_data, "undershoot-pct", 95, 0);
+        av_opt_set_int(encoderCtx, "drop-frame", 25, AV_OPT_SEARCH_CHILDREN);
+        av_opt_set_int(encoderCtx, "undershoot-pct", 95, AV_OPT_SEARCH_CHILDREN);
         // don't set encoderCtx->gop_size: let libvpx decide when to insert a keyframe
         encoderCtx->slices = 2; // VP8E_SET_TOKEN_PARTITIONS
         encoderCtx->qmin = 4;
@@ -244,7 +272,7 @@ MediaEncoder::addStream(const SystemCodecInfo& systemCodecInfo, std::string para
         encoderCtx->rc_buffer_size = maxBitrate;
         encoderCtx->bit_rate = maxBitrate;
         if (crf != SystemCodecInfo::DEFAULT_NO_QUALITY) {
-            av_opt_set_int(encoderCtx->priv_data, "crf", crf, 0);
+            av_opt_set_int(encoderCtx, "crf", crf, AV_OPT_SEARCH_CHILDREN);
             RING_DBG("Using quality factor %d", crf);
         } else {
             RING_DBG("Using Max bitrate %d", maxBitrate);
@@ -269,7 +297,8 @@ MediaEncoder::addStream(const SystemCodecInfo& systemCodecInfo, std::string para
 
     currentStreamIdx_ = stream->index;
 
-    if (avcodec_open2(encoderCtx, outputCodec, nullptr) < 0)
+    readConfig(&options_, encoderCtx);
+    if (avcodec_open2(encoderCtx, outputCodec, &options_) < 0)
         throw MediaEncoderException("Could not open encoder");
 
 #ifndef _WIN32
@@ -284,7 +313,15 @@ MediaEncoder::addStream(const SystemCodecInfo& systemCodecInfo, std::string para
         // allocate buffers for both scaled (pre-encoder) and encoded frames
         const int width = encoderCtx->width;
         const int height = encoderCtx->height;
-        const int format = encoderCtx->pix_fmt;
+        int format = encoderCtx->pix_fmt;
+#ifdef RING_ACCEL
+        if (accel_) {
+            // hardware encoders require a specific pixel format
+            auto desc = av_pix_fmt_desc_get(encoderCtx->pix_fmt);
+            if (desc && (desc->flags & AV_PIX_FMT_FLAG_HWACCEL))
+                format = accel_->getSoftwareFormat();
+        }
+#endif
         scaledFrameBufferSize_ = videoFrameSize(format, width, height);
         if (scaledFrameBufferSize_ < 0)
             throw MediaEncoderException(("Could not compute buffer size: " + libav_utils::getError(scaledFrameBufferSize_)).c_str());
@@ -299,20 +336,11 @@ MediaEncoder::addStream(const SystemCodecInfo& systemCodecInfo, std::string para
     return stream->index;
 }
 
-void MediaEncoder::setInterruptCallback(int (*cb)(void*), void *opaque)
-{
-    if (cb) {
-        outputCtx_->interrupt_callback.callback = cb;
-        outputCtx_->interrupt_callback.opaque = opaque;
-    } else {
-        outputCtx_->interrupt_callback.callback = 0;
-    }
-}
-
-void MediaEncoder::setIOContext(const std::unique_ptr<MediaIOHandle> &ioctx)
+void
+MediaEncoder::setIOContext(AVIOContext* ioctx)
 {
     if (ioctx) {
-        outputCtx_->pb = ioctx->getContext();
+        outputCtx_->pb = ioctx;
         outputCtx_->packet_size = outputCtx_->pb->buffer_size;
     } else {
         int ret = 0;
@@ -323,8 +351,9 @@ void MediaEncoder::setIOContext(const std::unique_ptr<MediaIOHandle> &ioctx)
 #endif
         if (!(outputCtx_->oformat->flags & AVFMT_NOFILE)) {
             if ((ret = avio_open(&outputCtx_->pb, filename, AVIO_FLAG_WRITE)) < 0) {
-                throw MediaEncoderException(
-                    std::string("Could not set IO context" + libav_utils::getError(ret)).c_str());
+                std::stringstream ss;
+                ss << "Could not set IO context for '" << filename << "': " << libav_utils::getError(ret);
+                throw MediaEncoderException(ss.str().c_str());
             }
         }
     }
@@ -357,7 +386,11 @@ MediaEncoder::encode(VideoFrame& input, bool is_keyframe,
 
     scaler_.scale_with_aspect(input, scaledFrame_);
 
-    auto frame = scaledFrame_.pointer();
+    // Copy frame so the VideoScaler can still use the software frame (input)
+    VideoFrame copy;
+    copy.copyFrom(scaledFrame_);
+
+    auto frame = copy.pointer();
     AVCodecContext* enc = encoders_[currentStreamIdx_];
     // ideally, time base is the inverse of framerate, but this may not always be the case
     if (enc->framerate.num == enc->time_base.den && enc->framerate.den == enc->time_base.num)
@@ -373,11 +406,25 @@ MediaEncoder::encode(VideoFrame& input, bool is_keyframe,
         frame->key_frame = 0;
     }
 
+#ifdef RING_ACCEL
+    // NOTE needs to be at same scope as call to encode
+    std::unique_ptr<VideoFrame> framePtr;
+    if (accel_) {
+        framePtr = accel_->transfer(copy);
+        if (!framePtr) {
+            RING_ERR() << "Hardware encoding failure";
+            return -1;
+        }
+        frame = framePtr->pointer();
+    }
+#endif
+
     return encode(frame, currentStreamIdx_);
 }
 #endif // RING_VIDEO
 
-int MediaEncoder::encodeAudio(AudioFrame& frame)
+int
+MediaEncoder::encodeAudio(AudioFrame& frame)
 {
     frame.pointer()->pts = sent_samples;
     sent_samples += frame.pointer()->nb_samples;
@@ -409,19 +456,7 @@ MediaEncoder::encode(AVFrame* frame, int streamIdx)
         }
 
         if (pkt.size) {
-            pkt.stream_index = streamIdx;
-            if (pkt.pts != AV_NOPTS_VALUE)
-                pkt.pts = av_rescale_q(pkt.pts, encoderCtx->time_base,
-                                       outputCtx_->streams[streamIdx]->time_base);
-            if (pkt.dts != AV_NOPTS_VALUE)
-                pkt.dts = av_rescale_q(pkt.dts, encoderCtx->time_base,
-                                       outputCtx_->streams[streamIdx]->time_base);
-
-            // write the compressed frame
-            ret = av_write_frame(outputCtx_, &pkt);
-            if (ret < 0) {
-                RING_ERR() << "av_write_frame failed: " << libav_utils::getError(ret);
-            } else
+            if (send(pkt, streamIdx))
                 break;
         }
     }
@@ -430,6 +465,27 @@ MediaEncoder::encode(AVFrame* frame, int streamIdx)
     return 0;
 }
 
+bool
+MediaEncoder::send(AVPacket& pkt, int streamIdx)
+{
+    if (streamIdx < 0)
+        streamIdx = currentStreamIdx_;
+    auto encoderCtx = encoders_[streamIdx];
+    pkt.stream_index = streamIdx;
+    if (pkt.pts != AV_NOPTS_VALUE)
+        pkt.pts = av_rescale_q(pkt.pts, encoderCtx->time_base,
+                               outputCtx_->streams[streamIdx]->time_base);
+    if (pkt.dts != AV_NOPTS_VALUE)
+        pkt.dts = av_rescale_q(pkt.dts, encoderCtx->time_base,
+                               outputCtx_->streams[streamIdx]->time_base);
+    // write the compressed frame
+    auto ret = av_write_frame(outputCtx_, &pkt);
+    if (ret < 0) {
+        RING_ERR() << "av_write_frame failed: " << libav_utils::getError(ret);
+    }
+    return ret >= 0;
+}
+
 int
 MediaEncoder::flush()
 {
@@ -468,14 +524,12 @@ MediaEncoder::print_sdp()
     return result;
 }
 
-AVCodecContext* MediaEncoder::prepareEncoderContext(AVCodec* outputCodec, bool is_video)
+AVCodecContext*
+MediaEncoder::prepareEncoderContext(AVCodec* outputCodec, bool is_video)
 {
     AVCodecContext* encoderCtx = avcodec_alloc_context3(outputCodec);
 
-    auto encoderName = encoderCtx->av_class->item_name ?
-        encoderCtx->av_class->item_name(encoderCtx) : nullptr;
-    if (encoderName == nullptr)
-        encoderName = "encoder?";
+    auto encoderName = outputCodec->name; // guaranteed to be non null if AVCodec is not null
 
     encoderCtx->thread_count = std::min(std::thread::hardware_concurrency(), is_video ? 16u : 4u);
     RING_DBG("[%s] Using %d threads", encoderName, encoderCtx->thread_count);
@@ -493,17 +547,16 @@ AVCodecContext* MediaEncoder::prepareEncoderContext(AVCodec* outputCodec, bool i
         // satisfy ffmpeg: denominator must be 16bit or less value
         // time base = 1/FPS
         if (device_.framerate) {
-            av_dict_set(&options_, "width", ring::to_string(device_.width).c_str(), 0);
             av_reduce(&encoderCtx->framerate.num, &encoderCtx->framerate.den,
                       device_.framerate.numerator(), device_.framerate.denominator(),
                       (1U << 16) - 1);
             encoderCtx->time_base = av_inv_q(encoderCtx->framerate);
         } else {
             // get from options_, else default to 30 fps
-            auto v = av_dict_get(options_, "framerate", nullptr, 0);
+            auto v = libav_utils::getDictValue(options_, "framerate");
             AVRational framerate = AVRational{30, 1};
             if (v)
-                av_parse_ratio_quiet(&framerate, v->value, 120);
+                av_parse_ratio_quiet(&framerate, v, 120);
             if (framerate.den == 0)
                 framerate.den = 1;
             av_reduce(&encoderCtx->framerate.num, &encoderCtx->framerate.den,
@@ -514,7 +567,11 @@ AVCodecContext* MediaEncoder::prepareEncoderContext(AVCodec* outputCodec, bool i
 
         // emit one intra frame every gop_size frames
         encoderCtx->max_b_frames = 0;
-        encoderCtx->pix_fmt = AV_PIX_FMT_YUV420P; // TODO: option me !
+        encoderCtx->pix_fmt = AV_PIX_FMT_YUV420P;
+#ifdef RING_ACCEL
+        if (accel_)
+            encoderCtx->pix_fmt = accel_->getFormat();
+#endif
 
         // Fri Jul 22 11:37:59 EDT 2011:tmatth:XXX: DON'T set this, we want our
         // pps and sps to be sent in-band for RTP
@@ -523,18 +580,18 @@ AVCodecContext* MediaEncoder::prepareEncoderContext(AVCodec* outputCodec, bool i
         // encoderCtx->flags |= AV_CODEC_FLAG_GLOBAL_HEADER;
     } else {
         encoderCtx->sample_fmt = AV_SAMPLE_FMT_S16;
-        auto v = av_dict_get(options_, "sample_rate", nullptr, 0);
+        auto v = libav_utils::getDictValue(options_, "sample_rate");
         if (v) {
-            encoderCtx->sample_rate = atoi(v->value);
+            encoderCtx->sample_rate = atoi(v);
             encoderCtx->time_base = AVRational{1, encoderCtx->sample_rate};
         } else {
             RING_WARN("[%s] No sample rate set", encoderName);
             encoderCtx->sample_rate = 8000;
         }
 
-        v = av_dict_get(options_, "channels", nullptr, 0);
+        v = libav_utils::getDictValue(options_, "channels");
         if (v) {
-            auto c = std::atoi(v->value);
+            auto c = std::atoi(v);
             if (c > 2 or c < 1) {
                 RING_WARN("[%s] Clamping invalid channel value %d", encoderName, c);
                 c = 1;
@@ -547,9 +604,9 @@ AVCodecContext* MediaEncoder::prepareEncoderContext(AVCodec* outputCodec, bool i
 
         encoderCtx->channel_layout = encoderCtx->channels == 2 ? AV_CH_LAYOUT_STEREO : AV_CH_LAYOUT_MONO;
 
-        v = av_dict_get(options_, "frame_size", nullptr, 0);
+        v = libav_utils::getDictValue(options_, "frame_size");
         if (v) {
-            encoderCtx->frame_size = atoi(v->value);
+            encoderCtx->frame_size = atoi(v);
             RING_DBG("[%s] Frame size %d", encoderName, encoderCtx->frame_size);
         } else {
             RING_WARN("[%s] Frame size not set", encoderName);
@@ -559,17 +616,19 @@ AVCodecContext* MediaEncoder::prepareEncoderContext(AVCodec* outputCodec, bool i
     return encoderCtx;
 }
 
-void MediaEncoder::forcePresetX264(AVCodecContext* encoderCtx)
+void
+MediaEncoder::forcePresetX264(AVCodecContext* encoderCtx)
 {
     const char *speedPreset = "ultrafast";
-    if (av_opt_set(encoderCtx->priv_data, "preset", speedPreset, 0))
+    if (av_opt_set(encoderCtx, "preset", speedPreset, AV_OPT_SEARCH_CHILDREN))
         RING_WARN("Failed to set x264 preset '%s'", speedPreset);
     const char *tune = "zerolatency";
-    if (av_opt_set(encoderCtx->priv_data, "tune", tune, 0))
+    if (av_opt_set(encoderCtx, "tune", tune, AV_OPT_SEARCH_CHILDREN))
         RING_WARN("Failed to set x264 tune '%s'", tune);
 }
 
-void MediaEncoder::extractProfileLevelID(const std::string &parameters,
+void
+MediaEncoder::extractProfileLevelID(const std::string &parameters,
                                          AVCodecContext *ctx)
 {
     // From RFC3984:
@@ -617,18 +676,26 @@ void MediaEncoder::extractProfileLevelID(const std::string &parameters,
     RING_DBG("Using profile %x and level %d", ctx->profile, ctx->level);
 }
 
-void
-MediaEncoder::setMuted(bool isMuted)
-{
-    is_muted = isMuted;
-}
-
 bool
 MediaEncoder::useCodec(const ring::AccountCodecInfo* codec) const noexcept
 {
     return codec_.get() == codec;
 }
 
+#ifdef RING_ACCEL
+void
+MediaEncoder::enableAccel(bool enableAccel)
+{
+    enableAccel_ = enableAccel;
+    emitSignal<DRing::ConfigurationSignal::HardwareEncodingChanged>(enableAccel_);
+    if (!enableAccel_) {
+        accel_.reset();
+        for (auto enc : encoders_)
+            enc->opaque = nullptr;
+    }
+}
+#endif
+
 unsigned
 MediaEncoder::getStreamCount() const
 {
@@ -649,7 +716,76 @@ MediaEncoder::getStream(const std::string& name, int streamIdx) const
         return {};
     auto enc = encoders_[streamIdx];
     // TODO set firstTimestamp
-    return MediaStream(name, enc);
+    auto ms = MediaStream(name, enc);
+#ifdef RING_ACCEL
+    if (accel_)
+        ms.format = accel_->getSoftwareFormat();
+#endif
+    return ms;
+}
+
+void
+MediaEncoder::readConfig(AVDictionary** dict, AVCodecContext* encoderCtx)
+{
+    std::string path = fileutils::get_config_dir() + DIR_SEPARATOR_STR + "encoder.json";
+    std::string name = encoderCtx->codec->name;
+    if (fileutils::isFile(path)) {
+        try {
+            Json::Value root;
+            std::ifstream file(path);
+            file >> root;
+            if (!root.isObject()) {
+                RING_ERR() << "Invalid encoder configuration: root is not an object";
+                return;
+            }
+            const auto& config = root[name];
+            if (config.isNull()) {
+                RING_WARN() << "Encoder '" << name << "' not found in configuration file";
+                return;
+            }
+            if (!config.isObject()) {
+                RING_ERR() << "Invalid encoder configuration: '" << name << "' is not an object";
+                return;
+            }
+            // If users want to change these, they should use the settings page.
+            std::vector<std::string> ignoredKeys = { "width", "height", "framerate", "sample_rate", "channels", "frame_size", "parameters" };
+            for (Json::Value::const_iterator it = config.begin(); it != config.end(); ++it) {
+                Json::Value v = *it;
+                if (!it.key().isConvertibleTo(Json::ValueType::stringValue)
+                    || !v.isConvertibleTo(Json::ValueType::stringValue)) {
+                    RING_ERR() << "Invalid configuration for '" << name << "'";
+                    return;
+                }
+                const auto& key = it.key().asString();
+                const auto& value = v.asString();
+                // provides a way to override all AVCodecContext fields MediaEncoder sets
+                if (std::find(ignoredKeys.cbegin(), ignoredKeys.cend(), key) != ignoredKeys.cend())
+                    continue;
+                else if (value.empty())
+                    libav_utils::setDictValue(dict, key, nullptr);
+                else if (key == "profile")
+                    encoderCtx->profile = v.asInt();
+                else if (key == "level")
+                    encoderCtx->level = v.asInt();
+                else if (key == "bit_rate")
+                    encoderCtx->bit_rate = v.asInt();
+                else if (key == "rc_buffer_size")
+                    encoderCtx->rc_buffer_size = v.asInt();
+                else if (key == "rc_min_rate")
+                    encoderCtx->rc_min_rate = v.asInt();
+                else if (key == "rc_max_rate")
+                    encoderCtx->rc_max_rate = v.asInt();
+                else if (key == "qmin")
+                    encoderCtx->qmin = v.asInt();
+                else if (key == "qmax")
+                    encoderCtx->qmax = v.asInt();
+                else
+                    libav_utils::setDictValue(dict, key, value);
+            }
+        } catch (const Json::Exception& e) {
+            RING_ERR() << "Failed to load encoder configuration file: " << e.what();
+        }
+    }
 }
 
 } // namespace ring
