@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2012-2018 Savoir-faire Linux Inc.
+ *  Copyright (C) 2012-2019 Savoir-faire Linux Inc.
  *
  *  Author: Tristan Matthews <tristan.matthews@savoirfairelinux.com>
  *  Author: Guillaume Roguez <guillaume.roguez@savoirfairelinux.com>
@@ -38,6 +38,11 @@
 #include "libav_utils.h"
 #include "video_scaler.h"
 #include "smartools.h"
+#include "media_filter.h"
+
+#ifdef RING_ACCEL
+#include "accel.h"
+#endif
 
 #ifndef _WIN32
 #include <sys/mman.h>
@@ -50,9 +55,16 @@
 #include <cerrno>
 #include <cstring>
 #include <stdexcept>
+#include <cmath>
+
+extern "C" {
+#include <libavutil/display.h>
+}
 
 namespace ring { namespace video {
 
+const constexpr char FILTER_INPUT_NAME[] = "in";
+
 #if HAVE_SHM
 // RAII class helper on sem_wait/sem_post sempahore operations
 class SemGuardLock {
@@ -78,14 +90,14 @@ class SemGuardLock {
 class ShmHolder
 {
     public:
-        ShmHolder(const std::string& name={});
+        ShmHolder(const std::string& name = {});
         ~ShmHolder();
 
         std::string name() const noexcept {
             return openedName_;
         }
 
-        void renderFrame(VideoFrame& src) noexcept;
+        void renderFrame(const VideoFrame& src) noexcept;
 
     private:
         bool resizeArea(std::size_t desired_length) noexcept;
@@ -220,7 +232,7 @@ ShmHolder::resizeArea(std::size_t frameSize) noexcept
 }
 
 void
-ShmHolder::renderFrame(VideoFrame& src) noexcept
+ShmHolder::renderFrame(const VideoFrame& src) noexcept
 {
     const auto width = src.width();
     const auto height = src.height();
@@ -312,12 +324,60 @@ SinkClient::SinkClient(const std::string& id, bool mixer)
 #endif
 {}
 
+void
+SinkClient::setRotation(int rotation)
+{
+    if (rotation_ == rotation || width_ == 0 || height_ == 0)
+        return;
+
+    rotation_ = rotation;
+    RING_WARN("Rotation set to %d", rotation_);
+    auto in_name = FILTER_INPUT_NAME;
+
+    std::stringstream ss;
+
+    ss << "[" << in_name << "] " << "format=rgb32,";  // avoid https://trac.ffmpeg.org/ticket/5356
+
+    switch (rotation_) {
+        case 90 :
+        case -270 :
+            ss << "transpose=2";
+            break;
+        case 180 :
+        case -180 :
+            ss << "transpose=2,transpose=2";
+            break;
+        case 270 :
+        case -90 :
+            ss << "transpose=1";
+            break;
+        default :
+            ss << "null";
+    }
+
+    const auto format = AV_PIX_FMT_RGB32;
+    const auto one = rational<int>(1);
+    std::vector<MediaStream> msv;
+    msv.emplace_back(in_name, format, one, width_, height_, one, one);
+
+    if (!rotation_) {
+        filter_.reset();
+    }
+    else {
+        filter_.reset(new MediaFilter);
+        auto ret = filter_->initialize(ss.str(), msv);
+        if (ret < 0) {
+            RING_ERR() << "filter init fail";
+            filter_ = nullptr;
+            rotation_ = 0;
+        }
+    }
+}
+
 void
 SinkClient::update(Observable<std::shared_ptr<MediaFrame>>* /*obs*/,
                    const std::shared_ptr<MediaFrame>& frame_p)
 {
-    auto& f = *std::static_pointer_cast<VideoFrame>(frame_p);
-
 #ifdef DEBUG_FPS
     auto currentTime = std::chrono::system_clock::now();
     const std::chrono::duration<double> seconds = currentTime - lastFrameDebug_;
@@ -332,31 +392,60 @@ SinkClient::update(Observable<std::shared_ptr<MediaFrame>>* /*obs*/,
     }
 #endif
 
+    if (avTarget_.push) {
+        auto outFrame = std::make_unique<VideoFrame>();
+        outFrame->copyFrom(*std::static_pointer_cast<VideoFrame>(frame_p));
+        avTarget_.push(std::move(outFrame));
+    }
+
+    bool doTransfer = (target_.pull != nullptr);
 #if HAVE_SHM
-    shm_->renderFrame(f);
+    doTransfer |= (shm_ != nullptr);
 #endif
 
-    if (target_.pull) {
-        VideoFrame dst;
-        VideoScaler scaler;
-        const int width = f.width();
-        const int height = f.height();
-#if defined(__ANDROID__) || (defined(__APPLE__) && !TARGET_OS_IPHONE)
-        const int format = AV_PIX_FMT_RGBA;
+    if (doTransfer) {
+#ifdef RING_ACCEL
+        std::shared_ptr<VideoFrame> frame {HardwareAccel::transferToMainMemory(*std::static_pointer_cast<VideoFrame>(frame_p), AV_PIX_FMT_NV12)};
 #else
-        const int format = AV_PIX_FMT_BGRA;
+        std::shared_ptr<VideoFrame> frame {std::static_pointer_cast<VideoFrame>(frame_p)};
 #endif
-
-        const auto bytes = videoFrameSize(format, width, height);
-
-        if (bytes > 0) {
-            if (auto buffer_ptr = target_.pull(bytes)) {
-                buffer_ptr->format = format;
-                buffer_ptr->width = width;
-                buffer_ptr->height = height;
-                dst.setFromMemory(buffer_ptr->ptr, format, width, height);
-                scaler_->scale(f, dst);
-                target_.push(std::move(buffer_ptr));
+        AVFrameSideData* side_data = av_frame_get_side_data(frame->pointer(), AV_FRAME_DATA_DISPLAYMATRIX);
+        if (side_data) {
+            auto matrix_rotation = reinterpret_cast<int32_t*>(side_data->data);
+            auto angle = av_display_rotation_get(matrix_rotation);
+            if (!std::isnan(angle))
+                setRotation(angle);
+            if (filter_) {
+                filter_->feedInput(frame->pointer(), FILTER_INPUT_NAME);
+                frame = std::static_pointer_cast<VideoFrame>(std::shared_ptr<MediaFrame>(filter_->readOutput()));
+            }
+            if (frame->height() != height_ || frame->width() != width_) {
+                setFrameSize(0, 0);
+                setFrameSize(frame->width(), frame->height());
+            }
+        }
+#if HAVE_SHM
+        shm_->renderFrame(*frame);
+#endif
+        if (target_.pull) {
+            VideoFrame dst;
+            const int width = frame->width();
+            const int height = frame->height();
+#if defined(__ANDROID__) || (defined(__APPLE__) && !TARGET_OS_IPHONE)
+            const int format = AV_PIX_FMT_RGBA;
+#else
+            const int format = AV_PIX_FMT_BGRA;
+#endif
+            const auto bytes = videoFrameSize(format, width, height);
+            if (bytes > 0) {
+                if (auto buffer_ptr = target_.pull(bytes)) {
+                    buffer_ptr->format = format;
+                    buffer_ptr->width = width;
+                    buffer_ptr->height = height;
+                    dst.setFromMemory(buffer_ptr->ptr, format, width, height);
+                    scaler_->scale(*frame, dst);
+                    target_.push(std::move(buffer_ptr));
+                }
             }
         }
     }
