@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2004-2016 Savoir-faire Linux Inc.
+ *  Copyright (C) 2004-2019 Savoir-faire Linux Inc.
  *
  *  Author: Alexandre Savard <alexandre.savard@savoirfairelinux.com>
  *
@@ -38,8 +38,15 @@
 #include "audio/pulseaudio/pulselayer.h"
 #endif
 #if HAVE_COREAUDIO
-#include "audio/coreaudio/corelayer.h"
+#ifdef __APPLE__
+#include <TargetConditionals.h>
 #endif
+#if TARGET_OS_IOS
+#include "audio/coreaudio/ios/corelayer.h"
+#else
+#include "audio/coreaudio/osx/corelayer.h"
+#endif /* TARGET_OS_IOS */
+#endif /* HAVE_COREAUDIO */
 #if HAVE_PORTAUDIO
 #include "audio/portaudio/portaudiolayer.h"
 #endif
@@ -72,75 +79,81 @@ const char * const Preferences::DFT_ZONE = "North America";
 const char * const Preferences::REGISTRATION_EXPIRE_KEY = "registrationexpire";
 
 // general preferences
-static const char * const ORDER_KEY = "order";
-static const char * const AUDIO_API_KEY = "audioApi";
-static const char * const HISTORY_LIMIT_KEY = "historyLimit";
-static const char * const HISTORY_MAX_CALLS_KEY = "historyMaxCalls";
-static const char * const ZONE_TONE_CHOICE_KEY = "zoneToneChoice";
-static const char * const PORT_NUM_KEY = "portNum";
-static const char * const SEARCH_BAR_DISPLAY_KEY = "searchBarDisplay";
-static const char * const MD5_HASH_KEY = "md5Hash";
+static constexpr const char* ORDER_KEY {"order"};
+static constexpr const char* AUDIO_API_KEY {"audioApi"};
+static constexpr const char* HISTORY_LIMIT_KEY {"historyLimit"};
+static constexpr const char* RINGING_TIMEOUT {"ringingTimeout"};
+static constexpr const char* HISTORY_MAX_CALLS_KEY {"historyMaxCalls"};
+static constexpr const char* ZONE_TONE_CHOICE_KEY {"zoneToneChoice"};
+static constexpr const char* PORT_NUM_KEY {"portNum"};
+static constexpr const char* SEARCH_BAR_DISPLAY_KEY {"searchBarDisplay"};
+static constexpr const char* MD5_HASH_KEY {"md5Hash"};
 
 // voip preferences
 constexpr const char * const VoipPreference::CONFIG_LABEL;
-static const char * const PLAY_DTMF_KEY = "playDtmf";
-static const char * const PLAY_TONES_KEY = "playTones";
-static const char * const PULSE_LENGTH_KEY = "pulseLength";
-static const char * const SYMMETRIC_RTP_KEY = "symmetric";
-static const char * const ZID_FILE_KEY = "zidFile";
+static constexpr const char* PLAY_DTMF_KEY {"playDtmf"};
+static constexpr const char* PLAY_TONES_KEY {"playTones"};
+static constexpr const char* PULSE_LENGTH_KEY {"pulseLength"};
+static constexpr const char* SYMMETRIC_RTP_KEY {"symmetric"};
+static constexpr const char* ZID_FILE_KEY {"zidFile"};
 
 // hooks preferences
 constexpr const char * const HookPreference::CONFIG_LABEL;
-static const char * const NUMBER_ADD_PREFIX_KEY = "numberAddPrefix";
-static const char * const NUMBER_ENABLED_KEY = "numberEnabled";
-static const char * const SIP_ENABLED_KEY = "sipEnabled";
-static const char * const URL_COMMAND_KEY = "urlCommand";
-static const char * const URL_SIP_FIELD_KEY = "urlSipField";
+static constexpr const char* NUMBER_ADD_PREFIX_KEY {"numberAddPrefix"};
+static constexpr const char* NUMBER_ENABLED_KEY {"numberEnabled"};
+static constexpr const char* SIP_ENABLED_KEY {"sipEnabled"};
+static constexpr const char* URL_COMMAND_KEY {"urlCommand"};
+static constexpr const char* URL_SIP_FIELD_KEY {"urlSipField"};
 
 // audio preferences
 constexpr const char * const AudioPreference::CONFIG_LABEL;
-static const char * const ALSAMAP_KEY = "alsa";
-static const char * const PULSEMAP_KEY = "pulse";
-static const char * const CARDIN_KEY = "cardIn";
-static const char * const CARDOUT_KEY = "cardOut";
-static const char * const CARDRING_KEY = "cardRing";
-static const char * const PLUGIN_KEY = "plugin";
-static const char * const SMPLRATE_KEY = "smplRate";
-static const char * const DEVICE_PLAYBACK_KEY = "devicePlayback";
-static const char * const DEVICE_RECORD_KEY = "deviceRecord";
-static const char * const DEVICE_RINGTONE_KEY = "deviceRingtone";
-static const char * const RECORDPATH_KEY = "recordPath";
-static const char * const ALWAYS_RECORDING_KEY = "alwaysRecording";
-static const char * const VOLUMEMIC_KEY = "volumeMic";
-static const char * const VOLUMESPKR_KEY = "volumeSpkr";
-static const char * const NOISE_REDUCE_KEY = "noiseReduce";
-static const char * const AGC_KEY = "automaticGainControl";
-static const char * const CAPTURE_MUTED_KEY = "captureMuted";
-static const char * const PLAYBACK_MUTED_KEY = "playbackMuted";
+static constexpr const char* ALSAMAP_KEY {"alsa"};
+static constexpr const char* PULSEMAP_KEY {"pulse"};
+static constexpr const char* CARDIN_KEY {"cardIn"};
+static constexpr const char* CARDOUT_KEY {"cardOut"};
+static constexpr const char* CARDRING_KEY {"cardRing"};
+static constexpr const char* PLUGIN_KEY {"plugin"};
+static constexpr const char* SMPLRATE_KEY {"smplRate"};
+static constexpr const char* DEVICE_PLAYBACK_KEY {"devicePlayback"};
+static constexpr const char* DEVICE_RECORD_KEY {"deviceRecord"};
+static constexpr const char* DEVICE_RINGTONE_KEY {"deviceRingtone"};
+static constexpr const char* RECORDPATH_KEY {"recordPath"};
+static constexpr const char* ALWAYS_RECORDING_KEY {"alwaysRecording"};
+static constexpr const char* VOLUMEMIC_KEY {"volumeMic"};
+static constexpr const char* VOLUMESPKR_KEY {"volumeSpkr"};
+static constexpr const char* NOISE_REDUCE_KEY {"noiseReduce"};
+static constexpr const char* AGC_KEY {"automaticGainControl"};
+static constexpr const char* CAPTURE_MUTED_KEY {"captureMuted"};
+static constexpr const char* PLAYBACK_MUTED_KEY {"playbackMuted"};
 
 // shortcut preferences
 constexpr const char * const ShortcutPreferences::CONFIG_LABEL;
-static const char * const HANGUP_SHORT_KEY = "hangUp";
-static const char * const PICKUP_SHORT_KEY = "pickUp";
-static const char * const POPUP_SHORT_KEY = "popupWindow";
-static const char * const TOGGLE_HOLD_SHORT_KEY = "toggleHold";
-static const char * const TOGGLE_PICKUP_HANGUP_SHORT_KEY = "togglePickupHangup";
+static constexpr const char* HANGUP_SHORT_KEY {"hangUp"};
+static constexpr const char* PICKUP_SHORT_KEY {"pickUp"};
+static constexpr const char* POPUP_SHORT_KEY {"popupWindow"};
+static constexpr const char* TOGGLE_HOLD_SHORT_KEY {"toggleHold"};
+static constexpr const char* TOGGLE_PICKUP_HANGUP_SHORT_KEY {"togglePickupHangup"};
 
 #ifdef RING_VIDEO
 // video preferences
 constexpr const char * const VideoPreferences::CONFIG_LABEL;
-#ifdef RING_ACCEL
-static const char * const DECODING_ACCELERATED_KEY = "decodingAccelerated";
-#endif
+static constexpr const char* DECODING_ACCELERATED_KEY {"decodingAccelerated"};
+static constexpr const char* ENCODING_ACCELERATED_KEY {"encodingAccelerated"};
 #endif
 
-static const char * const DFT_PULSE_LENGTH_STR = "250"; /** Default DTMF lenght */
-static const char * const ALSA_DFT_CARD    = "0";          /** Default sound card index */
+static constexpr const char* DFT_PULSE_LENGTH_STR {"250"}; /** Default DTMF length */
+#ifndef _MSC_VER
+static constexpr const char* ALSA_DFT_CARD {"0"};          /** Default sound card index */
+#else
+static constexpr const char* ALSA_DFT_CARD {"-1"};         /** Default sound card index (Portaudio) */
+#endif // _MSC_VER
+
 
 Preferences::Preferences() :
     accountOrder_("")
-    , historyLimit_(30)
+    , historyLimit_(0)
     , historyMaxCalls_(20)
+    , ringingTimeout_(30)
     , zoneToneChoice_(DFT_ZONE) // DFT_ZONE
     , registrationExpire_(180)
     , portNum_(sip_utils::DEFAULT_SIP_PORT)
@@ -197,6 +210,7 @@ void Preferences::serialize(YAML::Emitter &out)
     out << YAML::Key << CONFIG_LABEL << YAML::Value << YAML::BeginMap;
 
     out << YAML::Key << HISTORY_LIMIT_KEY << YAML::Value << historyLimit_;
+    out << YAML::Key << RINGING_TIMEOUT << YAML::Value << ringingTimeout_;
     out << YAML::Key << HISTORY_MAX_CALLS_KEY << YAML::Value << historyMaxCalls_;
     out << YAML::Key << MD5_HASH_KEY << YAML::Value << md5Hash_;
     out << YAML::Key << ORDER_KEY << YAML::Value << accountOrder_;
@@ -213,6 +227,7 @@ void Preferences::unserialize(const YAML::Node &in)
 
     parseValue(node, ORDER_KEY, accountOrder_);
     parseValue(node, HISTORY_LIMIT_KEY, historyLimit_);
+    parseValue(node, RINGING_TIMEOUT, ringingTimeout_);
     parseValue(node, HISTORY_MAX_CALLS_KEY, historyMaxCalls_);
     parseValue(node, ZONE_TONE_CHOICE_KEY, zoneToneChoice_);
     parseValue(node, REGISTRATION_EXPIRE_KEY, registrationExpire_);
@@ -547,9 +562,8 @@ void ShortcutPreferences::unserialize(const YAML::Node &in)
 
 #ifdef RING_VIDEO
 VideoPreferences::VideoPreferences()
-#ifdef RING_ACCEL
-    : decodingAccelerated_(false)
-#endif
+    : decodingAccelerated_(true)
+    , encodingAccelerated_(false)
 {
 }
 
@@ -558,6 +572,7 @@ void VideoPreferences::serialize(YAML::Emitter &out)
     out << YAML::Key << CONFIG_LABEL << YAML::Value << YAML::BeginMap;
 #ifdef RING_ACCEL
     out << YAML::Key << DECODING_ACCELERATED_KEY << YAML::Value << decodingAccelerated_;
+    out << YAML::Key << ENCODING_ACCELERATED_KEY << YAML::Value << encodingAccelerated_;
 #endif
     getVideoDeviceMonitor().serialize(out);
     out << YAML::EndMap;
@@ -570,7 +585,11 @@ void VideoPreferences::unserialize(const YAML::Node &in)
     // value may or may not be present
     try {
         parseValue(node, DECODING_ACCELERATED_KEY, decodingAccelerated_);
-    } catch (...) { decodingAccelerated_ = false; } // experimental, so disabled by default
+        parseValue(node, ENCODING_ACCELERATED_KEY, encodingAccelerated_);
+    } catch (...) {
+        decodingAccelerated_ = true;
+        encodingAccelerated_ = false;
+    }
 #endif
     getVideoDeviceMonitor().unserialize(in);
 }
