@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2004-2018 Savoir-faire Linux Inc.
+ *  Copyright (C) 2004-2019 Savoir-faire Linux Inc.
  *
  *  Author: Adrien Béraud <adrien.beraud@savoirfairelinux.com>
  *  Author: Guillaume Roguez <guillaume.roguez@savoirfairelinux.com>
@@ -31,6 +31,7 @@
 #include "certstore.h"
 #include "array_size.h"
 #include "diffie-hellman.h"
+#include "scheduled_executor.h"
 
 #include <gnutls/gnutls.h>
 #include <gnutls/dtls.h>
@@ -56,10 +57,11 @@ static constexpr const char* DTLS_CERT_PRIORITY_STRING {"SECURE192:-VERS-TLS-ALL
 static constexpr const char* DTLS_FULL_PRIORITY_STRING {"SECURE192:-KX-ALL:+ANON-ECDH:+ANON-DH:+SECURE192:-VERS-TLS-ALL:+VERS-DTLS-ALL:-RSA:%SERVER_PRECEDENCE:%SAFE_RENEGOTIATION"};
 static constexpr const char* TLS_CERT_PRIORITY_STRING {"SECURE192:-RSA:%SERVER_PRECEDENCE:%SAFE_RENEGOTIATION"};
 static constexpr const char* TLS_FULL_PRIORITY_STRING {"SECURE192:-KX-ALL:+ANON-ECDH:+ANON-DH:+SECURE192:-RSA:%SERVER_PRECEDENCE:%SAFE_RENEGOTIATION"};
-static constexpr uint16_t INPUT_BUFFER_SIZE {16*1024}; // to be coherent with the maximum size advised in path mtu discovery
+static constexpr uint32_t RX_MAX_SIZE {64*1024}; // 64k = max size of a UDP packet
 static constexpr std::size_t INPUT_MAX_SIZE {1000}; // Maximum number of packets to store before dropping (pkt size = DTLS_MTU)
 static constexpr ssize_t FLOOD_THRESHOLD {4*1024};
 static constexpr auto FLOOD_PAUSE = std::chrono::milliseconds(100); // Time to wait after an invalid cookie packet (anti flood attack)
+static constexpr size_t HANDSHAKE_MAX_RETRY {64};
 static constexpr auto DTLS_RETRANSMIT_TIMEOUT = std::chrono::milliseconds(1000); // Delay between two handshake request on DTLS
 static constexpr auto COOKIE_TIMEOUT = std::chrono::seconds(10); // Time to wait for a cookie packet from client
 static constexpr int MIN_MTU {512 - 20 - 8}; // minimal payload size of a DTLS packet carried by an IPv4 packet
@@ -247,6 +249,11 @@ public:
     bool setup();
     void process();
     void cleanup();
+    // State protectors
+    std::mutex stateMutex_;
+    std::condition_variable stateCondition_;
+
+    ScheduledExecutor scheduler_;
 
     // Path mtu discovery
     std::array<int, 3> MTUS_;
@@ -287,8 +294,6 @@ TlsSession::TlsSessionImpl::TlsSessionImpl(SocketType& transport,
             });
     }
 
-    Manager::instance().registerEventHandler((uintptr_t)this, [this]{ flushRxQueue(); });
-
     // Run FSM into dedicated thread
     thread_.start();
 }
@@ -298,7 +303,6 @@ TlsSession::TlsSessionImpl::~TlsSessionImpl()
     thread_.join();
     if (not transport_.isReliable())
         transport_.setOnRecv(nullptr);
-    Manager::instance().unregisterEventHandler((uintptr_t)this);
 }
 
 const char*
@@ -671,11 +675,15 @@ TlsSession::TlsSessionImpl::waitForRawData(unsigned timeout)
 
     // non-reliable uses callback installed with setOnRecv()
     std::unique_lock<std::mutex> lk {rxMutex_};
-    rxCv_.wait(lk, [this]{ return !rxQueue_.empty() or state_ == TlsSessionState::SHUTDOWN; });
+    rxCv_.wait_for(lk, std::chrono::milliseconds(timeout), [this]{ return !rxQueue_.empty() or state_ == TlsSessionState::SHUTDOWN; });
     if (state_ == TlsSessionState::SHUTDOWN) {
         gnutls_transport_set_errno(session_, EINTR);
         return -1;
     }
+    if (rxQueue_.empty()) {
+        RING_ERR("[TLS] waitForRawData: timeout after %u ms", timeout);
+        return 0;
+    }
     return 1;
 }
 
@@ -713,6 +721,7 @@ void
 TlsSession::TlsSessionImpl::cleanup()
 {
     state_ = TlsSessionState::SHUTDOWN; // be sure to block any user operations
+    stateCondition_.notify_all();
 
     if (session_) {
         if (transport_.isReliable())
@@ -822,9 +831,14 @@ TlsSession::TlsSessionImpl::handleStateCookie(TlsSessionState state)
 TlsSessionState
 TlsSession::TlsSessionImpl::handleStateHandshake(TlsSessionState state)
 {
-    RING_DBG("[TLS] handshake");
-
-    auto ret = gnutls_handshake(session_);
+    int ret;
+    size_t retry_count = 0;
+    do {
+        RING_DBG("[TLS] handshake");
+        ret = gnutls_handshake(session_);
+    } while ((ret == GNUTLS_E_INTERRUPTED   or
+              ret == GNUTLS_E_AGAIN       ) and
+              ++retry_count < HANDSHAKE_MAX_RETRY);
 
     // Stop on fatal error
     if (gnutls_error_is_fatal(ret)) {
@@ -841,16 +855,26 @@ TlsSession::TlsSessionImpl::handleStateHandshake(TlsSessionState state)
     }
 
     // Safe-Renegotiation status shall always be true to prevent MiM attack
-    if (!gnutls_safe_renegotiation_status(session_)) {
-        RING_ERR("[TLS] server identity changed! MiM attack?");
-        return TlsSessionState::SHUTDOWN;
+    // Following https://www.gnutls.org/manual/html_node/Safe-renegotiation.html
+    // "Unlike TLS 1.2, the server is not allowed to change identities"
+    // So, we don't have to check the status if we are the client
+#if GNUTLS_VERSION_NUMBER >= 0x030605
+    bool isTLS1_3 = gnutls_protocol_get_version(session_) == GNUTLS_TLS1_3;
+    if (!isTLS1_3 || (isTLS1_3 && isServer_)) {
+#endif
+        if (!gnutls_safe_renegotiation_status(session_)) {
+            RING_ERR("[TLS] server identity changed! MiM attack?");
+            return TlsSessionState::SHUTDOWN;
+        }
+#if GNUTLS_VERSION_NUMBER >= 0x030605
     }
+#endif
 
     auto desc = gnutls_session_get_desc(session_);
     RING_DBG("[TLS] session established: %s", desc);
     gnutls_free(desc);
 
-    // Anonymous connection? rehandshake immediatly with certificate authentification forced
+    // Anonymous connection? rehandshake immediately with certificate authentification forced
     auto cred = gnutls_auth_get_type(session_);
     if (cred == GNUTLS_CRD_ANON) {
         RING_DBG("[TLS] renogotiate with certificate authentification");
@@ -1022,12 +1046,13 @@ TlsSession::TlsSessionImpl::handleDataPacket(std::vector<ValueType>&& buf, uint6
 
     // Try to flush right now as a new packet is available
     flushRxQueue();
+    scheduler_.scheduleIn([this]{ flushRxQueue(); }, RX_OOO_TIMEOUT);
 }
 
 ///
 /// Reorder and push received packet to upper layer
 ///
-/// \note This method must be called continously, faster than RX_OOO_TIMEOUT
+/// \note This method must be called continuously, faster than RX_OOO_TIMEOUT
 ///
 void
 TlsSession::TlsSessionImpl::flushRxQueue()
@@ -1051,11 +1076,13 @@ TlsSession::TlsSessionImpl::flushRxQueue()
 
     GuardedBoolSwap swap_flush_processing {flushProcessing_};
 
+    auto now = clock::now();
+
     auto item = std::begin(reorderBuffer_);
     auto next_offset = item->first;
 
-    // Wait for next continous packet until timeout
-    if ((clock::now() - lastReadTime_) >= RX_OOO_TIMEOUT) {
+    // Wait for next continuous packet until timeout
+    if ((now - lastReadTime_) >= RX_OOO_TIMEOUT) {
         // OOO packet timeout - consider waited packets as lost
         if (auto lost = next_offset - gapOffset_)
             RING_WARN("[TLS] %lu lost since 0x%lx", lost, gapOffset_);
@@ -1081,7 +1108,7 @@ TlsSession::TlsSessionImpl::flushRxQueue()
     }
 
     gapOffset_ = std::max(gapOffset_, next_offset);
-    lastReadTime_ = clock::now();
+    lastReadTime_ = now;
 }
 
 TlsSessionState
@@ -1089,13 +1116,12 @@ TlsSession::TlsSessionImpl::handleStateEstablished(TlsSessionState state)
 {
     // Nothing to do in reliable mode, so just wait for state change
     if (transport_.isReliable()) {
-        while (true) {
-            std::this_thread::sleep_for(std::chrono::milliseconds(100));
-            state = state_.load();
-            if (state != TlsSessionState::ESTABLISHED)
-                return state;
-        }
-        return TlsSessionState::SHUTDOWN;
+        auto disconnected = [this]() ->  bool {
+            return state_.load() != TlsSessionState::ESTABLISHED;
+        };
+        std::unique_lock<std::mutex> lk(stateMutex_);
+        stateCondition_.wait(lk, disconnected);
+        return state;
     }
 
     // block until rx packet or state change
@@ -1108,7 +1134,7 @@ TlsSession::TlsSessionImpl::handleStateEstablished(TlsSessionState state)
     }
 
     std::array<uint8_t, 8> seq;
-    rawPktBuf_.resize(maxPayload_);
+    rawPktBuf_.resize(RX_MAX_SIZE);
     auto ret = gnutls_record_recv_seq(session_, rawPktBuf_.data(), rawPktBuf_.size(), &seq[0]);
 
     if (ret > 0) {
@@ -1172,6 +1198,9 @@ TlsSession::TlsSessionImpl::process()
     if (not std::atomic_compare_exchange_strong(&state_, &old_state, new_state))
         new_state = old_state;
 
+    if (old_state != new_state)
+        stateCondition_.notify_all();
+
     if (old_state != new_state and callbacks_.onStateChange)
         callbacks_.onStateChange(new_state);
 }
@@ -1237,6 +1266,7 @@ void
 TlsSession::shutdown()
 {
     pimpl_->state_ = TlsSessionState::SHUTDOWN;
+    pimpl_->stateCondition_.notify_all();
     pimpl_->rxCv_.notify_one(); // unblock waiting FSM
     pimpl_->transport_.shutdown();
 }
@@ -1278,6 +1308,7 @@ TlsSession::read(ValueType* data, std::size_t size, std::error_code& ec)
             RING_DBG("[TLS] re-handshake");
             pimpl_->state_ = TlsSessionState::HANDSHAKE;
             pimpl_->rxCv_.notify_one(); // unblock waiting FSM
+            pimpl_->stateCondition_.notify_all();
         } else if (gnutls_error_is_fatal(ret)) {
             RING_ERR("[TLS] fatal error in recv: %s", gnutls_strerror(ret));
             shutdown();
@@ -1291,13 +1322,20 @@ TlsSession::read(ValueType* data, std::size_t size, std::error_code& ec)
 }
 
 void
-TlsSession::connect()
+TlsSession::waitForReady(const std::chrono::steady_clock::duration& timeout)
 {
-    TlsSessionState state;
-    do {
-        state = pimpl_->state_.load();
-        std::this_thread::sleep_for(std::chrono::milliseconds(10));
-    } while (state != TlsSessionState::ESTABLISHED and state != TlsSessionState::SHUTDOWN);
+    auto ready = [this]() ->  bool {
+        auto state = pimpl_->state_.load();
+        return state == TlsSessionState::ESTABLISHED or state == TlsSessionState::SHUTDOWN;
+    };
+    std::unique_lock<std::mutex> lk(pimpl_->stateMutex_);
+    if (timeout == std::chrono::steady_clock::duration::zero())
+        pimpl_->stateCondition_.wait(lk, ready);
+    else
+        pimpl_->stateCondition_.wait_for(lk, timeout, ready);
+
+    if(!ready())
+        throw std::logic_error("Invalid state in TlsSession::waitForReady");
 }
 
 int
